#!/usr/bin/python
# dht_relations.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (dht_relations.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: dht_relations

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from twisted.internet.defer import Deferred, DeferredList  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import strng

from bitdust.dht import dht_records

from bitdust.contacts import contactsdb

from bitdust.userid import my_id
from bitdust.userid import id_url

from bitdust.contacts import identitycache

#------------------------------------------------------------------------------


def read_customer_suppliers(customer_idurl, as_fields=True, use_cache=True):
    if as_fields:
        customer_idurl = id_url.field(customer_idurl)
    else:
        customer_idurl = id_url.to_bin(customer_idurl)

    rotated_idurls = id_url.list_known_idurls(customer_idurl, num_revisions=3)

    if _Debug:
        lg.args(_DebugLevel, customer_idurl=customer_idurl, rotated_idurls=rotated_idurls, as_fields=as_fields, use_cache=use_cache)

    result = Deferred()

    def _do_identity_cache(ret):
        all_stories = []
        for _supplier_idurl in ret['suppliers']:
            if _supplier_idurl:
                _supplier_idurl = id_url.to_bin(_supplier_idurl)
                if not id_url.is_cached(_supplier_idurl) or not identitycache.HasFile(_supplier_idurl):
                    one_supplier_story = identitycache.immediatelyCaching(_supplier_idurl)
                    one_supplier_story.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='read_customer_suppliers._do_identity_cache')
                    all_stories.append(one_supplier_story)
        _customer_idurl = id_url.to_bin(ret['customer_idurl'])
        if _customer_idurl and (not id_url.is_cached(_customer_idurl) or not identitycache.HasFile(_customer_idurl)):
            one_customer_story = identitycache.immediatelyCaching(_customer_idurl)
            one_customer_story.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='read_customer_suppliers._do_identity_cache')
            all_stories.append(one_customer_story)
        if _Debug:
            lg.args(_DebugLevel, all_stories=len(all_stories), ret=ret)
        id_cache_story = DeferredList(all_stories, consumeErrors=True)
        id_cache_story.addCallback(_do_save_customer_suppliers, ret)
        id_cache_story.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='read_customer_suppliers._do_identity_cache')
        id_cache_story.addErrback(result.errback)
        return id_cache_story

    def _do_verify(dht_value, customer_idurl_bin):
        if customer_idurl_bin in rotated_idurls:
            rotated_idurls.remove(customer_idurl_bin)
        ret = {
            'suppliers': [],
            'ecc_map': None,
            'customer_idurl': customer_idurl,
            'revision': 0,
            'publisher_idurl': None,
            'timestamp': None,
        }
        if not dht_value or not isinstance(dht_value, dict):
            if not rotated_idurls:
                result.callback(ret)
                return ret
            another_customer_idurl_bin = rotated_idurls.pop(0)
            lg.warn('found another rotated idurl %r and re-try reading customer suppliers' % another_customer_idurl_bin)
            d = dht_records.get_suppliers(another_customer_idurl_bin, return_details=True, use_cache=False)
            d.addCallback(_do_verify, another_customer_idurl_bin)
            d.addErrback(_on_error)
            return ret
        try:
            _ecc_map = strng.to_text(dht_value['ecc_map'])
            if as_fields:
                _customer_idurl = id_url.field(dht_value['customer_idurl'])
                _publisher_idurl = id_url.field(dht_value.get('publisher_idurl'))
                _suppliers_list = id_url.fields_list(dht_value['suppliers'])
            else:
                _customer_idurl = id_url.to_bin(dht_value['customer_idurl'])
                _publisher_idurl = id_url.to_bin(dht_value.get('publisher_idurl'))
                _suppliers_list = id_url.to_bin_list(dht_value['suppliers'])
            _revision = int(dht_value.get('revision'))
            _timestamp = int(dht_value.get('timestamp'))
        except:
            lg.exc()
            result.callback(ret)
            return ret
        ret.update({
            'suppliers': _suppliers_list,
            'ecc_map': _ecc_map,
            'customer_idurl': _customer_idurl,
            'revision': _revision,
            'publisher_idurl': _publisher_idurl,
            'timestamp': _timestamp,
        })
        return _do_identity_cache(ret)

    def _do_save_customer_suppliers(id_cached_result, ret):
        if my_id.getIDURL() != id_url.field(ret['customer_idurl']):
            contactsdb.set_suppliers(ret['suppliers'], customer_idurl=ret['customer_idurl'])
            contactsdb.save_suppliers(customer_idurl=ret['customer_idurl'])
            if ret.get('ecc_map'):
                for supplier_idurl in ret['suppliers']:
                    if supplier_idurl and id_url.is_cached(supplier_idurl):
                        contactsdb.add_supplier_meta_info(
                            supplier_idurl=supplier_idurl,
                            info={
                                'ecc_map': ret['ecc_map'],
                            },
                            customer_idurl=ret['customer_idurl'],
                        )
        else:
            if _Debug:
                lg.out(_DebugLevel, 'dht_relations._do_save_customer_suppliers SKIP processing my own suppliers')
        if _Debug:
            lg.out(_DebugLevel, 'dht_relations._do_save_customer_suppliers  OK  for %r  returned %d suppliers' % (ret['customer_idurl'], len(ret['suppliers'])))
        result.callback(ret)
        return ret

    def _on_error(err):
        try:
            msg = err.getErrorMessage()
        except:
            msg = str(err).replace('Exception:', '')
        if _Debug:
            lg.out(_DebugLevel, 'dht_relations.read_customer_suppliers ERROR %r  failed with %r' % (customer_idurl, msg))
        result.errback(err)
        return None

    customer_idurl_bin = id_url.to_bin(customer_idurl)
    d = dht_records.get_suppliers(customer_idurl_bin, return_details=True, use_cache=use_cache)
    d.addCallback(_do_verify, customer_idurl_bin)
    d.addErrback(_on_error)
    return result


def write_customer_suppliers(
    customer_idurl,
    suppliers_list,
    ecc_map=None,
    revision=None,
    publisher_idurl=None,
):
    customer_idurl = id_url.field(customer_idurl)
    publisher_idurl = id_url.field(publisher_idurl)
    if customer_idurl == my_id.getIDURL():
        lg.warn('skip writing my own suppliers list which suppose to be written to DHT')
    else:
        contactsdb.set_suppliers(suppliers_list, customer_idurl=customer_idurl)
        contactsdb.save_suppliers(customer_idurl=customer_idurl)
    return dht_records.set_suppliers(
        customer_idurl=customer_idurl,
        suppliers_list=id_url.fields_list(suppliers_list),
        ecc_map=ecc_map,
        revision=revision,
        publisher_idurl=publisher_idurl,
    )


#------------------------------------------------------------------------------


def read_customer_message_brokers(customer_idurl, positions=[
    0,
], return_details=True, as_fields=True, use_cache=True):
    if _Debug:
        lg.args(_DebugLevel, customer_idurl=customer_idurl, use_cache=use_cache, positions=positions)
    if as_fields:
        customer_idurl = id_url.field(customer_idurl)
    else:
        customer_idurl = id_url.to_bin(customer_idurl)
    result = Deferred()

    def _on_borker_identity_cache_failed(err, position, broker_result):
        if _Debug:
            lg.args(_DebugLevel, position=position, err=err)
        broker_result.callback({
            'timestamp': None,
            'revision': 0,
            'customer_idurl': customer_idurl,
            'broker_idurl': None,
            'position': position,
        })
        return None

    def _do_broker_identity_cache(dht_record, position, broker_result):
        one_broker_task = identitycache.GetLatest(id_url.to_bin(dht_record['broker_idurl']))
        one_broker_task.addCallback(lambda xmlsrc: broker_result.callback(dht_record))
        one_broker_task.addErrback(_on_borker_identity_cache_failed, position, broker_result)
        # if _Debug:
        #     lg.args(_DebugLevel, position=position, broker_idurl=dht_record['broker_idurl'])
        return None

    def _do_verify(dht_value, position, broker_result):
        ret = {
            'timestamp': None,
            'revision': 0,
            'customer_idurl': customer_idurl,
            'broker_idurl': None,
            'position': position,
        }
        if not dht_value or not isinstance(dht_value, dict):
            if _Debug:
                lg.args(_DebugLevel, c=customer_idurl, p=position, dht_value=type(dht_value))
            broker_result.callback(ret)
            return ret
        try:
            if as_fields:
                _customer_idurl = id_url.field(dht_value['customer_idurl'])
                _broker_idurl = id_url.field(dht_value['broker_idurl'])
            else:
                _customer_idurl = id_url.to_bin(dht_value['customer_idurl'])
                _broker_idurl = id_url.to_bin(dht_value['broker_idurl'])
            _position = int(dht_value['position'])
            _revision = int(dht_value.get('revision'))
            _timestamp = int(dht_value.get('timestamp'))
        except:
            lg.exc()
            broker_result.callback(ret)
            return ret
        if _Debug:
            lg.args(_DebugLevel, p=position, b=_broker_idurl, r=_revision)
        if as_fields:
            if _customer_idurl != customer_idurl:
                lg.err('wrong customer idurl %r in message broker DHT record for %r at position %d' % (_customer_idurl, customer_idurl, position))
                broker_result.callback(ret)
                return ret
        if position != _position:
            lg.err('wrong position value %d in message broker DHT record for %r at position %d' % (_position, customer_idurl, position))
            broker_result.callback(ret)
            return ret
        ret.update({
            'customer_idurl': _customer_idurl,
            'broker_idurl': _broker_idurl,
            'position': _position,
            'revision': _revision,
            'timestamp': _timestamp,
        })
        _do_broker_identity_cache(ret, position, broker_result)
        return None

    def _on_error(err, position, broker_result):
        try:
            msg = err.getErrorMessage()
        except:
            msg = str(err).replace('Exception:', '')
        if _Debug:
            lg.out(_DebugLevel, 'dht_relations.read_customer_message_brokers ERROR %r at position %d failed with %r' % (customer_idurl, position, msg))
        broker_result.errback(err)
        return None

    def _do_collect_results(all_results):
        # if _Debug:
        #     lg.args(_DebugLevel, all_results=len(all_results))
        final_result = []
        for one_success, one_result in all_results:
            if one_success and one_result['broker_idurl']:
                final_result.append(one_result)
        final_result.sort(key=lambda i: i.get('position'))
        if _Debug:
            lg.args(_DebugLevel, results=len(final_result))
        result.callback(final_result)
        return None

    def _do_read_brokers():
        all_brokers_results = []
        for position in positions:
            one_broker_result = Deferred()
            all_brokers_results.append(one_broker_result)
            d = dht_records.get_message_broker(
                customer_idurl=customer_idurl,
                position=position,
                return_details=return_details,
                use_cache=use_cache,
            )
            d.addCallback(_do_verify, position, one_broker_result)
            d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='read_customer_message_brokers._do_read_brokers')
            d.addErrback(_on_error, position, one_broker_result)
        join_all_brokers = DeferredList(all_brokers_results, consumeErrors=False)
        join_all_brokers.addCallback(_do_collect_results)
        join_all_brokers.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='read_customer_message_brokers._do_read_brokers')
        join_all_brokers.addErrback(result.errback)
        return None

    d = identitycache.GetLatest(customer_idurl)
    d.addCallback(lambda _: _do_read_brokers())
    d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='read_customer_message_brokers')
    d.addErrback(result.errback)
    return result


def write_customer_message_broker(customer_idurl, broker_idurl, position=0, revision=None):
    if _Debug:
        lg.args(_DebugLevel, c=customer_idurl, b=broker_idurl, p=position, r=revision)
    customer_idurl = id_url.field(customer_idurl)
    broker_idurl = id_url.field(broker_idurl)
    return dht_records.set_message_broker(
        customer_idurl=customer_idurl,
        broker_idurl=broker_idurl,
        position=position,
        revision=revision,
    )



#!/usr/bin/python
# service_p2p_notifications.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_p2p_notifications.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_p2p_notifications
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return P2PNotificationsService()


class P2PNotificationsService(LocalService):

    service_name = 'service_p2p_notifications'
    config_path = 'services/p2p-notifications/enabled'

    def dependent_on(self):
        from bitdust.main import settings
        depends = [
            'service_p2p_hookups',
        ]
        if settings.enablePROXY():
            depends.append('service_proxy_transport')
        return depends

    def start(self):
        from twisted.internet.task import LoopingCall
        from bitdust.transport import callback
        from bitdust.stream import p2p_queue
        p2p_queue.init()
        callback.append_inbox_callback(self._on_inbox_packet_received)
        self.reconnect_task = LoopingCall(self._on_check_network_connect)
        self.reconnect_task.start(30, now=False)
        return True

    def stop(self):
        from bitdust.transport import callback
        from bitdust.stream import p2p_queue
        if self.reconnect_task and self.reconnect_task.running:
            self.reconnect_task.stop()
        self.reconnect_task = None
        callback.remove_inbox_callback(self._on_inbox_packet_received)
        p2p_queue.shutdown()
        return True

    def request(self, json_payload, newpacket, info):
        from bitdust.logs import lg
        from bitdust.lib import serialization
        from bitdust.p2p import p2p_service
        from bitdust.stream import p2p_queue
        try:
            service_requests_list = json_payload['items']
        except:
            lg.exc()
            return p2p_service.SendFail(newpacket, 'invalid payload')
        service_responses_list = []
        for r_json in service_requests_list:
            resp = r_json.copy()
            r_scope = r_json.get('scope', '')
            r_action = r_json.get('action', '')
            try:
                if r_scope == 'queue':
                    if r_action == 'open':
                        resp['result'] = 'denied' if not p2p_queue.open_queue(queue_id=r_json.get('queue_id')) else 'OK'
                    elif r_action == 'close':
                        resp['result'] = 'denied' if not p2p_queue.close_queue(queue_id=r_json.get('queue_id')) else 'OK'
                elif r_scope == 'consumer':
                    if r_action == 'start':
                        resp['result'] = 'denied' if not p2p_queue.add_consumer(consumer_id=r_json.get('consumer_id')) else 'OK'
                    elif r_action == 'stop':
                        resp['result'] = 'denied' if not p2p_queue.remove_consumer(consumer_id=r_json.get('consumer_id')) else 'OK'
                    elif r_action == 'add_callback':
                        resp['result'] = 'denied' if not p2p_queue.add_callback_method(
                            consumer_id=r_json.get('consumer_id'),
                            callback_method=r_json.get('method'),
                            interested_queues_list=r_json.get('queues') or None,
                        ) else 'OK'
                    elif r_action == 'remove_callback':
                        resp['result'] = 'denied' if not p2p_queue.remove_callback_method(
                            consumer_id=r_json.get('consumer_id'),
                            callback_method=r_json.get('method'),
                        ) else 'OK'
                    elif r_action == 'subscribe':
                        resp['result'] = 'denied' if not p2p_queue.subscribe_consumer(
                            consumer_id=r_json.get('consumer_id'),
                            queue_id=r_json.get('queue_id'),
                        ) else 'OK'
                    elif r_action == 'unsubscribe':
                        resp['result'] = 'denied' if not p2p_queue.unsubscribe_consumer(
                            consumer_id=r_json.get('consumer_id'),
                            queue_id=r_json.get('queue_id'),
                        ) else 'OK'
                elif r_scope == 'producer':
                    resp['result'] = 'denied'
                    resp['reason'] = 'remote requests for producing messages is not allowed'
                    if False:
                        # TODO: do we need that ?
                        if r_action == 'start':
                            resp['result'] = 'denied' if not p2p_queue.add_producer(producer_id=r_json.get('producer_id')) else 'OK'
                        elif r_action == 'stop':
                            resp['result'] = 'denied' if not p2p_queue.remove_producer(producer_id=r_json.get('producer_id')) else 'OK'
                        elif r_action == 'connect':
                            resp['result'] = 'denied' if not p2p_queue.connect_producer(
                                producer_id=r_json.get('producer_id'),
                                queue_id=r_json.get('queue_id'),
                            ) else 'OK'
                        elif r_action == 'disconnect':
                            resp['result'] = 'denied' if not p2p_queue.disconnect_producer(
                                producer_id=r_json.get('producer_id'),
                                queue_id=r_json.get('queue_id'),
                            ) else 'OK'
            except Exception as exc:
                resp['result'] = 'denied'
                resp['reason'] = str(exc)
            service_responses_list.append(resp)
            lg.out(self.debug_level, 'service_p2p_notifications.request  %s:%s  is  [%s] : %s' % (r_scope, r_action, resp['result'], resp.get('reason', 'OK')))
        payload = serialization.DictToBytes({'items': service_responses_list}, values_to_text=True)
        return p2p_service.SendAck(newpacket, payload)

    def cancel(self, json_payload, newpacket, info):
        # TODO: work in progress
        return False

    def _on_inbox_packet_received(self, newpacket, info, status, error_message):
        from bitdust.p2p import commands
        from bitdust.stream import p2p_queue
        if newpacket.Command != commands.Event():
            return False
        return p2p_queue.on_event_packet_received(newpacket, info, status, error_message)

    def _on_check_network_connect(self):
        from bitdust.p2p import network_service
        from bitdust.userid import my_id
        if my_id.isLocalIdentityReady():
            network_service.connected(wait_timeout=0.1)
        return None



#!/usr/bin/python
# service_coins_accountant.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_accountant.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_coins_accountant
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return AccountantService()


class AccountantService(LocalService):

    service_name = 'service_accountant'
    config_path = 'services/accountant/enabled'

    def dependent_on(self):
        return [
            # 'service_broadcasting',
            # 'service_blockchain',
        ]

    def installed(self):
        # TODO: to be continue...
        return False

    def start(self):
        from coins import coins_db
        from coins import accountant_node
        from coins import accountants_finder
        coins_db.init()
        accountants_finder.A('init')
        accountant_node.A('init')
        accountant_node.A('start')
        accountant_node.A().addStateChangedCallback(self._on_accountant_node_switched)
        return True

    def stop(self):
        from coins import coins_db
        from coins import accountant_node
        accountant_node.A().removeStateChangedCallback(self._on_accountant_node_switched)
        accountant_node.A('stop')
        accountant_node.A('shutdown')
        coins_db.shutdown()
        return True

    def request(self, json_payload, newpacket, info):
        from bitdust.logs import lg
        from bitdust.p2p import p2p_service
        # words = newpacket.Payload.split(' ')
        try:
            # mode = words[1][:10]
            mode = json_payload['action']
        except:
            lg.exc()
            return p2p_service.SendFail(newpacket, 'invalid payload')
        if mode != 'join' and mode != 'write' and mode != 'read':
            lg.out(8, 'service_accountant.request DENIED, wrong mode provided : %s' % mode)
            return p2p_service.SendFail(newpacket, 'invalid request')
        from coins import accountant_node
        if not accountant_node.A():
            lg.out(8, 'service_accountant.request DENIED, accountant_node() state machine not exist')
            return p2p_service.SendFail(newpacket, 'accountant_node service not started')
        # if accountant_node.A().state not in ['ACCOUNTANTS?', "READY", "VALID_COIN?", "WRITE_COIN!", ]:
        #     lg.out(8, "service_accountant.request DENIED, accountant_node() state is : %s" % accountant_node.A().state)
        # return p2p_service.SendFail(request, "accountant_node service
        # currently unavailable")
        if mode == 'join':
            accountant_node.A('accountant-connected', newpacket.OwnerID)
            #             if accountant_node.A().state == 'OFFLINE':
            #                 accountant_node.A('start')
            return p2p_service.SendAck(newpacket, 'accepted')
        return p2p_service.SendFail(newpacket, 'bad request')

    def health_check(self):
        from coins import accountant_node
        return accountant_node.A().state in [
            'READY',
            'READ_COINS',
            'VALID_COIN?',
            'WRITE_COIN!',
        ]

    def _on_accountant_node_switched(self, oldstate, newstate, evt, *args, **kwargs):
        from bitdust.logs import lg
        from twisted.internet import reactor  # @UnresolvedImport
        from coins import accountant_node
        if newstate == 'OFFLINE' and oldstate != 'AT_STARTUP':
            reactor.callLater(10*60, accountant_node.A, 'start')  # @UndefinedVariable
            lg.out(8, 'service_broadcasting._on_accountant_node_switched will try to reconnect again after 10 minutes')



#!/usr/bin/env python
# payment.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (payment.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: payment.

"""

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import utime
from bitdust.lib import strng
from bitdust.lib import jsn

from bitdust.system import bpio
from bitdust.system import local_fs

from bitdust.main import settings

from bitdust.crypt import key

from bitdust.blockchain import bismuth_wallet

from bitdust.userid import id_url
from bitdust.userid import my_id

#------------------------------------------------------------------------------


def get_supplier_contracts_dir(supplier_idurl):
    supplier_idurl = id_url.field(supplier_idurl)
    supplier_contracts_prefix = '{}_{}'.format(
        supplier_idurl.username,
        strng.to_text(key.HashSHA(supplier_idurl.to_public_key(), hexdigest=True)),
    )
    return os.path.join(settings.ServiceDir('service_customer_contracts'), supplier_contracts_prefix)


def save_storage_contract(supplier_idurl, json_data):
    if _Debug:
        lg.args(_DebugLevel, s=supplier_idurl, json_data=json_data)
    supplier_contracts_dir = get_supplier_contracts_dir(supplier_idurl)
    if not os.path.isdir(supplier_contracts_dir):
        bpio._dirs_make(supplier_contracts_dir)
    contract_path = os.path.join(supplier_contracts_dir, str(utime.unpack_time(json_data['started'])))
    json_data['supplier'] = supplier_idurl
    local_fs.WriteTextFile(contract_path, jsn.dumps(json_data))
    return contract_path


def list_storage_contracts(supplier_idurl):
    supplier_contracts_dir = get_supplier_contracts_dir(supplier_idurl)
    if not os.path.isdir(supplier_contracts_dir):
        return []
    l = []
    for contract_filename in os.listdir(supplier_contracts_dir):
        contract_path = os.path.join(supplier_contracts_dir, contract_filename)
        json_data = jsn.loads_text(local_fs.ReadTextFile(contract_path))
        l.append(json_data)
    l.sort(key=lambda json_data: utime.unpack_time(json_data['started']))
    if _Debug:
        lg.args(_DebugLevel, s=supplier_idurl, l=len(l))
    return l


def list_contracted_suppliers():
    customer_contracts_dir = settings.ServiceDir('service_customer_contracts')
    if not os.path.isdir(customer_contracts_dir):
        bpio._dirs_make(customer_contracts_dir)
    supplier_idurls = []
    for supplier_contracts_prefix in os.listdir(customer_contracts_dir):
        supplier_contracts_dir = os.path.join(customer_contracts_dir, supplier_contracts_prefix)
        for contract_filename in os.listdir(supplier_contracts_dir):
            contract_path = os.path.join(supplier_contracts_dir, contract_filename)
            json_data = jsn.loads_text(local_fs.ReadTextFile(contract_path))
            supplier_idurl = json_data.get('supplier', None)
            if supplier_idurl:
                if not id_url.is_in(supplier_idurl, supplier_idurls):
                    supplier_idurls.append(supplier_idurl)
                break
    return supplier_idurls


#------------------------------------------------------------------------------


def pay_for_storage():
    cur_balance = bismuth_wallet.my_balance()
    if cur_balance == 'N/A':
        lg.err('my current balance is not available, payments are not possible at the moment')
        return False
    now = utime.utcnow_to_sec1970()
    my_customer_prefix = my_id.getIDURL().unique_name()
    if _Debug:
        lg.args(_DebugLevel, my_balance=cur_balance, my_customer_prefix=my_customer_prefix)
    for supplier_idurl in list_contracted_suppliers():
        supplier_contracts = list_storage_contracts(supplier_idurl)
        unpaid_contracts = []
        unpaid_sequence_numbers = set()
        for pos, json_data in enumerate(supplier_contracts):
            if json_data.get('paid'):
                continue
            if now < utime.unpack_time(json_data['complete_after']):
                continue
            unpaid_contracts.append(json_data)
            sequence_number = json_data.get('sequence_number', -1)
            if sequence_number < 0:
                sequence_number = pos + 1
            unpaid_sequence_numbers.add(sequence_number)
        sum_to_pay = 0
        pay_before_earliest = None
        pay_before_latest = None
        supplier_wallet_address = None
        started = None
        complete_after = None
        if unpaid_contracts:
            unpaid_contracts.sort(key=lambda i: utime.unpack_time(i['started']))
            started = unpaid_contracts[0]['started']
            complete_after = unpaid_contracts[-1]['complete_after']
            pay_before_earliest = utime.unpack_time(complete_after)
            pay_before_latest = utime.unpack_time(unpaid_contracts[-1]['pay_before'])
            supplier_wallet_address = unpaid_contracts[-1]['wallet_address']
            for unpaid_contract in unpaid_contracts:
                sum_to_pay += unpaid_contract['value']
        if _Debug:
            lg.args(
                _DebugLevel,
                s=supplier_idurl,
                unpaid=len(unpaid_contracts),
                debt=sum_to_pay,
                pay_earliest=utime.pack_time(pay_before_earliest),
                pay_latest=utime.pack_time(pay_before_latest),
                started=started,
                complete_after=complete_after,
            )
        if sum_to_pay:
            if now > pay_before_earliest:
                if sum_to_pay >= cur_balance:
                    lg.warn('my current balance %r is not enough to pay %r to %r' % (cur_balance, sum_to_pay, supplier_idurl))
                    continue
                try:
                    ret = bismuth_wallet.send_transaction(
                        recipient=supplier_wallet_address,
                        amount=sum_to_pay,
                        operation='storage',
                        data='{} {}'.format(my_customer_prefix, ','.join(map(str, sorted(unpaid_sequence_numbers)))),
                        raise_errors=True,
                    )
                except:
                    lg.exc()
                    return False
                if _Debug:
                    lg.args(_DebugLevel, recipient=supplier_wallet_address, ret=ret)
                cur_balance = bismuth_wallet.my_balance()
                if cur_balance == 'N/A':
                    lg.err('my current balance is not available, payments are not possible at the moment')
                    return False
    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
import functools
from heapq import nsmallest
from operator import itemgetter
from collections import defaultdict
import six

try:
    from collections import Counter
except ImportError:
    class Counter(dict):

        'Mapping where default values are zero'

        def __missing__(self, key):
            return 0


def cache1lvl(maxsize=100):
    """
    modified version of http://code.activestate.com/recipes/498245/
    """
    def decorating_function(user_function):
        cache = {}
        use_count = Counter()

        @functools.wraps(user_function)
        def wrapper(key, *args, **kwargs):
            try:
                result = cache[key]
            except KeyError:
                if len(cache) == maxsize:
                    for k, _ in nsmallest(maxsize // 10 or 1,
                                          six.iteritems(use_count),
                                          key=itemgetter(1)):
                        del cache[k], use_count[k]
                cache[key] = user_function(key, *args, **kwargs)
                result = cache[key]
                # result = user_function(obj, key, *args, **kwargs)
            finally:
                use_count[key] += 1
            return result

        def clear():
            cache.clear()
            use_count.clear()

        def delete(key):
            try:
                del cache[key]
                del use_count[key]
            except KeyError:
                return False
            else:
                return True

        wrapper.clear = clear
        wrapper.cache = cache
        wrapper.delete = delete
        return wrapper
    return decorating_function


def twolvl_iterator(dict):
    for k, v in six.iteritems(dict):
        for kk, vv in six.iteritems(v):
            yield k, kk, vv


def cache2lvl(maxsize=100):
    """
    modified version of http://code.activestate.com/recipes/498245/
    """
    def decorating_function(user_function):
        cache = {}
        use_count = defaultdict(Counter)

        @functools.wraps(user_function)
        def wrapper(*args, **kwargs):
#            return user_function(*args, **kwargs)
            try:
                result = cache[args[0]][args[1]]
            except KeyError:
                if wrapper.cache_size == maxsize:
                    to_delete = maxsize // 10 or 1
                    for k1, k2, v in nsmallest(to_delete,
                                               twolvl_iterator(use_count),
                                               key=itemgetter(2)):
                        del cache[k1][k2], use_count[k1][k2]
                        if not cache[k1]:
                            del cache[k1]
                            del use_count[k1]
                    wrapper.cache_size -= to_delete
                result = user_function(*args, **kwargs)
                try:
                    cache[args[0]][args[1]] = result
                except KeyError:
                    cache[args[0]] = {args[1]: result}
                wrapper.cache_size += 1
            finally:
                use_count[args[0]][args[1]] += 1
            return result

        def clear():
            cache.clear()
            use_count.clear()

        def delete(key, inner_key=None):
            if inner_key is not None:
                try:
                    del cache[key][inner_key]
                    del use_count[key][inner_key]
                    if not cache[key]:
                        del cache[key]
                        del use_count[key]
                    wrapper.cache_size -= 1
                except KeyError:
                    return False
                else:
                    return True
            else:
                try:
                    wrapper.cache_size -= len(cache[key])
                    del cache[key]
                    del use_count[key]
                except KeyError:
                    return False
                else:
                    return True

        wrapper.clear = clear
        wrapper.cache = cache
        wrapper.delete = delete
        wrapper.cache_size = 0
        return wrapper
    return decorating_function



#!/usr/bin/python
# listeners.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (listeners.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: listeners.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import sys

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in listeners.py')

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import utime

#------------------------------------------------------------------------------

_Listeners = {}
_ModelsToBePopulated = []

#------------------------------------------------------------------------------


def listeners():
    global _Listeners
    return _Listeners


#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'listeners.init')


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'listeners.shutdown')
    clear_listeners()


#------------------------------------------------------------------------------


class Snapshot(object):

    def __init__(self, model_name, snap_id=None, data=None, created=None, deleted=False):
        self.model_name = model_name
        self.snap_id = snap_id
        self.data = data
        self.created = created or utime.utcnow_to_sec1970()
        self.deleted = deleted

    def __repr__(self):
        return '[{}:{}{}>'.format(self.model_name, self.snap_id, ' DELETED' if self.deleted else '')

    def to_json(self):
        j = {
            'name': self.model_name,
            'id': self.snap_id,
            'data': self.data,
            'created': self.created,
        }
        if self.deleted:
            j['deleted'] = utime.utcnow_to_sec1970()
        return j


#------------------------------------------------------------------------------


def add_listener(listener_callback, model_name='*'):
    """
    listener_callback(snapshot_object)
    """
    if listener_callback in listeners().get(model_name, []):
        return False
    if model_name not in listeners():
        listeners()[model_name] = []
    listeners()[model_name].append(listener_callback)
    if _Debug:
        lg.args(_DebugLevel, m=model_name, cb=listener_callback)
    return True


def remove_listener(listener_callback, model_name='*'):
    removed = False
    if model_name == '*':
        for model_name, listeners_callbacks in listeners().items():
            if listener_callback in listeners_callbacks:
                listeners()[model_name].remove(listener_callback)
                removed = True
        if removed:
            for model_name in list(listeners().keys()):
                if not listeners()[model_name]:
                    listeners().pop(model_name)

    else:
        if model_name in listeners():
            if listener_callback in listeners()[model_name]:
                listeners()[model_name].remove(listener_callback)
                removed = True
                if not listeners()[model_name]:
                    listeners().pop(model_name)
    if _Debug:
        lg.args(_DebugLevel, m=model_name, cb=listener_callback, removed=removed)
    return removed


def clear_listeners(model_name='*'):
    removed = False
    total = 0
    for _model_name, listener_callbacks in listeners().items():
        if _model_name == model_name or model_name == '*':
            for _cb in list(listener_callbacks):
                listeners()[_model_name].remove(_cb)
                removed = True
                total += 1
    listeners().clear()
    if _Debug:
        lg.args(_DebugLevel, total=total)
    return removed


#------------------------------------------------------------------------------


def dispatch_snapshot(snap):
    handled = 0
    if snap.model_name in listeners():
        for listener_callback in listeners()[snap.model_name]:
            try:
                listener_callback(snap)
            except:
                lg.exc()
                continue
            handled += 1
    if '*' in listeners():
        for listener_callback in listeners()['*']:
            try:
                listener_callback(snap)
            except:
                lg.exc()
                continue
            handled += 1
    if _Debug:
        lg.args(_DebugLevel, handled=handled, snap=snap)
    return handled


#------------------------------------------------------------------------------


def push_snapshot(model_name, snap_id=None, data=None, created=None, deleted=False, fast=False):
    snap = Snapshot(model_name, snap_id=snap_id, data=data, created=created, deleted=deleted)
    # if _Debug:
    #     lg.args(_DebugLevel, s=snap, d=data)
    if fast:
        dispatch_snapshot(snap)
    else:
        reactor.callLater(0, dispatch_snapshot, snap)  # @UndefinedVariable
    return snap


#------------------------------------------------------------------------------


def populate_later(model_name=None, stop=False):
    global _ModelsToBePopulated
    if model_name is not None:
        if model_name not in _ModelsToBePopulated:
            _ModelsToBePopulated.append(model_name)
            if _Debug:
                lg.args(_DebugLevel, model_name=model_name)
        else:
            if stop:
                _ModelsToBePopulated.remove(model_name)
                if _Debug:
                    lg.args(_DebugLevel, model_name=model_name, stop=True)
    return _ModelsToBePopulated


def is_populate_required(model_name):
    return model_name in _ModelsToBePopulated



#!/usr/bin/env python
# customers_rejector.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (customers_rejector.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: customers_rejector.

.. role:: red

BitDust customers_rejector() Automat

.. raw:: html

    <a href="customers_rejector.png" target="_blank">
    <img src="customers_rejector.png" style="max-width:100%;">
    </a>

EVENTS:
    * :red:`contracts-verified`
    * :red:`customers-rejected`
    * :red:`found-idle-customers`
    * :red:`found-unpaid-customers`
    * :red:`no-idle-customers`
    * :red:`restart`
    * :red:`space-enough`
    * :red:`space-overflow`
    * :red:`start`
    * :red:`timer-1hour`
"""

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 8

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.system import bpio

from bitdust.main import settings
from bitdust.main import config

from bitdust.interface import api

from bitdust.contacts import contactsdb

from bitdust.lib import utime

from bitdust.services import driver

from bitdust.p2p import ratings

from bitdust.storage import accounting

#------------------------------------------------------------------------------

_CustomersRejector = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _CustomersRejector
    if _CustomersRejector is None:
        # set automat name and starting state here
        _CustomersRejector = CustomersRejector(
            name='customers_rejector',
            state='READY',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _CustomersRejector.automat(event, *args, **kwargs)
    return _CustomersRejector


def Destroy():
    """
    Destroy customers_rejector() automat and remove its instance from memory.
    """
    global _CustomersRejector
    if _CustomersRejector is None:
        return
    _CustomersRejector.destroy()
    del _CustomersRejector
    _CustomersRejector = None


class CustomersRejector(automat.Automat):

    """
    This class implements all the functionality of the ``customers_rejector()``
    state machine.
    """

    timers = {
        'timer-1hour': (3600, ['READY']),
    }

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """

    def A(self, event, *args, **kwargs):
        #---READY---
        if self.state == 'READY':
            if event == 'timer-1hour' or event == 'start' or event == 'restart':
                self.state = 'CAPACITY?'
                self.doTestMyCapacity(*args, **kwargs)
        #---CAPACITY?---
        elif self.state == 'CAPACITY?':
            if event == 'space-overflow':
                self.state = 'REJECT!'
                self.doRemoveCustomers(*args, **kwargs)
            elif event == 'space-enough':
                self.state = 'CONTRACTS?'
                self.doVerifyStorageContracts(*args, **kwargs)
        #---REJECT!---
        elif self.state == 'REJECT!':
            if event == 'restart':
                self.state = 'CAPACITY?'
                self.doTestMyCapacity(*args, **kwargs)
            elif event == 'customers-rejected':
                self.state = 'READY'
                self.doRestartLocalTester(*args, **kwargs)
                self.doRestartLater(*args, **kwargs)
        #---IDLE?---
        elif self.state == 'IDLE?':
            if event == 'found-idle-customers':
                self.state = 'REJECT!'
                self.doRemoveCustomers(*args, **kwargs)
            elif event == 'restart':
                self.state = 'CAPACITY?'
                self.doTestMyCapacity(*args, **kwargs)
            elif event == 'no-idle-customers':
                self.state = 'READY'
                self.doCountStatistics(*args, **kwargs)
        #---CONTRACTS?---
        elif self.state == 'CONTRACTS?':
            if event == 'found-unpaid-customers':
                self.state = 'REJECT!'
                self.doRemoveCustomers(*args, **kwargs)
            elif event == 'contracts-verified':
                self.state = 'IDLE?'
                self.doTestIdleDays(*args, **kwargs)
        return None

    def doTestMyCapacity(self, *args, **kwargs):
        """
        Here are some values.

        + donated_bytes : you set this in the settings
        + consumed_bytes : how many space was taken from you by other users
        + free_bytes = donated_bytes - consumed_bytes : not yet allocated space
        + used_bytes : size of all files, which you store on your disk for your customers
        + ratio : currently used space compared to consumed space
        """
        if _Debug:
            lg.out(_DebugLevel, 'customers_rejector.doTestMyCapacity')
        if bpio.Android():
            #TODO:
            self.automat('space-enough')
            return
        failed_customers = set()
        current_customers = contactsdb.customers()
        donated_bytes = settings.getDonatedBytes()
        space_dict, free_space = accounting.read_customers_quotas()
        used_dict = accounting.read_customers_usage()
        unknown_customers, unused_quotas = accounting.validate_customers_quotas(space_dict, free_space)
        failed_customers.update(unknown_customers)
        for idurl in unknown_customers:
            space_dict.pop(idurl, None)
        for idurl in unused_quotas:
            space_dict.pop(idurl, None)
        consumed_bytes = accounting.count_consumed_space(space_dict)
        free_space = donated_bytes - consumed_bytes
        if consumed_bytes < donated_bytes and len(failed_customers) == 0:
            accounting.write_customers_quotas(space_dict, free_space)
            lg.info('storage quota checks succeed, all customers are verified')
            self.automat('space-enough')
            return
        if failed_customers:
            for idurl in failed_customers:
                lg.warn('customer %r failed storage quota verification' % idurl)
                current_customers.remove(idurl)
            self.automat('space-overflow', failed_customers)
            return
        used_space_ratio_dict = accounting.calculate_customers_usage_ratio(space_dict, used_dict)
        customers_sorted = sorted(
            current_customers,
            key=lambda idurl: used_space_ratio_dict[idurl],
        )
        while len(customers_sorted) > 0 and consumed_bytes > donated_bytes:
            idurl = customers_sorted.pop()
            allocated_bytes = int(space_dict[idurl])
            consumed_bytes -= allocated_bytes
            space_dict.pop(idurl)
            failed_customers.add(idurl)
            current_customers.remove(idurl)
            lg.warn('customer %r will be removed because of storage quota overflow' % idurl)
        free_space = donated_bytes - consumed_bytes
        self.automat('space-overflow', failed_customers)

    def doTestIdleDays(self, *args, **kwargs):
        """
        Action method.
        """
        dead_customers = []
        customer_idle_days = config.conf().getInt('services/customer-patrol/customer-idle-days', 0)
        if not customer_idle_days:
            self.automat('no-idle-customers')
            return
        for customer_idurl in contactsdb.customers():
            if driver.is_on('service_supplier_contracts'):
                from bitdust.supplier import storage_contract
                if storage_contract.is_current_customer_contract_active(customer_idurl):
                    continue
            connected_time = ratings.connected_time(customer_idurl.to_bin())
            if connected_time is None:
                lg.warn('last connected_time for customer %r is unknown, rejecting customer' % customer_idurl)
                dead_customers.append(customer_idurl)
                continue
            if utime.utcnow_to_sec1970() - connected_time > customer_idle_days*24*60*60:
                lg.warn('customer %r connected last time %r seconds ago, rejecting customer' % (customer_idurl, utime.utcnow_to_sec1970() - connected_time))
                dead_customers.append(customer_idurl)
        if dead_customers:
            lg.warn('found idle customers: %r' % dead_customers)
            self.automat('found-idle-customers', dead_customers)
        else:
            lg.info('all customers has some activity recently, no idle customers found')
            self.automat('no-idle-customers')

    def doRestartLocalTester(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.supplier import local_tester
        local_tester.TestSpaceTime()

    def doRestartLater(self, *args, **kwargs):
        """
        Action method.
        """
        self.automat('restart', delay=5)

    def doVerifyStorageContracts(self, *args, **kwargs):
        """
        Action method.
        """
        rejected_customers = []
        if driver.is_on('service_supplier_contracts'):
            from bitdust.supplier import storage_contract
            rejected_customers = storage_contract.verify_all_current_customers_contracts()
        if rejected_customers:
            lg.warn('found unpaid customers: %r' % rejected_customers)
            # TODO: disabled for now...
            # self.automat('found-unpaid-customers', rejected_customers)
            self.automat('contracts-verified')
        else:
            lg.info('all customers have valid contracts')
            self.automat('contracts-verified')

    def doCountStatistics(self, *args, **kwargs):
        """
        Action method.
        """

    def doRemoveCustomers(self, *args, **kwargs):
        """
        Action method.
        """
        removed_customers = args[0]
        for customer_idurl in removed_customers:
            api.customer_reject(customer_id=customer_idurl, erase_customer_key=True)
        self.automat('customers-rejected', removed_customers)



#!/usr/bin/python
# service_private_messages.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_private_messages.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_private_messages
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return PrivateMessagesService()


class PrivateMessagesService(LocalService):

    service_name = 'service_private_messages'
    config_path = 'services/private-messages/enabled'
    start_suspended = True

    def dependent_on(self):
        return [
            'service_keys_registry',
            'service_entangled_dht',
        ]

    def start(self):
        from bitdust.main import events
        from bitdust.transport import callback
        from bitdust.stream import message
        from bitdust.chat import nickname_holder
        message.init()
        nickname_holder.A('set')
        callback.append_inbox_callback(self._on_inbox_packet_received)
        events.add_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        events.add_subscriber(self._on_user_connected, 'node-connected')
        events.add_subscriber(self._on_user_disconnected, 'node-disconnected')
        return True

    def stop(self):
        from bitdust.main import events
        from bitdust.transport import callback
        from bitdust.stream import message
        from bitdust.chat import nickname_holder
        events.remove_subscriber(self._on_user_connected, 'node-connected')
        events.remove_subscriber(self._on_user_disconnected, 'node-disconnected')
        events.remove_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        callback.remove_inbox_callback(self._on_inbox_packet_received)
        nickname_holder.Destroy()
        message.shutdown()
        return True

    def on_suspend(self, *args, **kwargs):
        return True

    def on_resume(self, *args, **kwargs):
        from bitdust.chat import nickname_holder
        nickname_holder.A('set')
        return True

    def _on_inbox_packet_received(self, newpacket, info, status, error_message):
        from bitdust.p2p import commands
        from bitdust.stream import message
        if newpacket.Command != commands.Message():
            return False
        return message.on_incoming_message(newpacket, info, status, error_message)

    def _on_identity_url_changed(self, evt):
        from bitdust.logs import lg
        from bitdust.contacts import contactsdb
        from bitdust.userid import id_url
        old_idurl = id_url.field(evt.data['old_idurl'])
        new_idurl = id_url.field(evt.data['new_idurl'])
        contacts_changed = False
        for idurl, alias in list(contactsdb.correspondents()):
            if old_idurl.to_bin() == id_url.field(idurl).to_bin():
                contactsdb.remove_correspondent(idurl)
                contactsdb.add_correspondent(new_idurl.to_bin(), alias)
                contacts_changed = True
                lg.info('found correspondent idurl was rotated : %r -> %r' % (old_idurl, new_idurl))
        if contacts_changed:
            contactsdb.save_correspondents()

    def _on_user_connected(self, evt):
        from bitdust.main import events
        from bitdust.contacts import contactsdb
        if contactsdb.is_correspondent(evt.data['idurl']):
            events.send('friend-connected', data=dict(
                idurl=evt.data['idurl'],
                global_id=evt.data['global_id'],
                old_state=evt.data['old_state'],
                new_state=evt.data['new_state'],
            ))

    def _on_user_disconnected(self, evt):
        from bitdust.main import events
        from bitdust.contacts import contactsdb
        if contactsdb.is_correspondent(evt.data['idurl']):
            events.send('friend-disconnected', data=dict(
                idurl=evt.data['idurl'],
                global_id=evt.data['global_id'],
                old_state=evt.data['old_state'],
                new_state=evt.data['new_state'],
            ))



#!/usr/bin/env python
# initializer.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (initializer.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: initializer.

.. raw:: html

    <a href="https://bitdust.io/automats/initializer/initializer.png" target="_blank">
    <img src="https://bitdust.io/automats/initializer/initializer.png" style="max-width:100%;">
    </a>

This Automat is the "entry point" to run all other state machines.
It manages the process of initialization of the whole program.

It also checks whether the program is installed and switch to run another "installation" code if needed.

The ``initializer()`` machine is doing several operations:

    * start low-level modules and init local data, see ``initializer.init_local()``
    * starts the network communications by running core method ``initializer.doInitServices()``
    * other modules is started after all other more important things
    * machine can switch to "install" wizard if Private Key or local identity file is not fine
    * it will finally switch to "READY" state to indicate the whole status of the application
    * during shutting down it will wait for ``shutdowner()`` automat to do its work completely
    * once ``shutdowner()`` become "FINISHED" - the state machine changes its state to "EXIT" and is destroyed


EVENTS:
    * :red:`init-interfaces-done`
    * :red:`init-local-done`
    * :red:`init-modules-done`
    * :red:`init-services-done`
    * :red:`installer.state`
    * :red:`run`
    * :red:`run-cmd-line-recover`
    * :red:`run-cmd-line-register`
    * :red:`shutdowner.state`
"""

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import os
import sys

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in initializer.py')

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import local_fs

from bitdust.crypt import cipher

from bitdust.automats import automat

from bitdust.services import driver

#------------------------------------------------------------------------------

_Initializer = None

#------------------------------------------------------------------------------


def init_settings(base_dir=None, network_name=None, override_configs=None, enable_debug=False):
    from bitdust.main import settings
    from bitdust.main import config
    if override_configs:
        settings.override_dict(override_configs)
    settings.init(base_dir=base_dir, network_name=network_name)
    if enable_debug:
        lg.set_debug_level(settings.getDebugLevel())
    config.conf().addConfigNotifier('logs/debug-level', lambda p, value, o, r: lg.set_debug_level(value))


def init_engine():
    from bitdust.contacts import identitydb
    from bitdust.userid import id_url
    from bitdust.main import listeners
    from bitdust.main import events
    events.init()
    listeners.init()
    id_url.init()
    identitydb.init()


def init_automats():
    from bitdust.main import config
    automat.init()
    automat.LifeBegins(lg.when_life_begins())
    automat.SetGlobalLogEvents(config.conf().getBool('logs/automat-events-enabled'))
    automat.SetGlobalLogTransitions(config.conf().getBool('logs/automat-transitions-enabled'))
    automat.SetExceptionsHandler(lg.exc)
    automat.SetLogOutputHandler(lambda debug_level, message: lg.out(debug_level, message, log_name='state'))
    # automat.OpenLogFile(settings.AutomatsLog())


def init_local():
    from bitdust.main import settings
    from bitdust.p2p import commands
    from bitdust.lib import net_misc
    from bitdust.lib import misc
    from bitdust.system import tmpfile
    from bitdust.system import run_upnpc
    from bitdust.raid import eccmap
    from bitdust.crypt import my_keys
    from bitdust.userid import my_id
    my_id.init()
    if settings.enableWebStream():
        try:
            from bitdust.logs import weblog
            weblog.init(settings.getWebStreamPort())
        except:
            lg.exc()
    if settings.enableWebTraffic():
        try:
            from bitdust.logs import webtraffic
            webtraffic.init(port=settings.getWebTrafficPort())
        except:
            lg.exc()
    misc.init()
    commands.init()
    tmpfile.init(settings.getTempDir())
    net_misc.init()
    settings.update_proxy_settings()
    run_upnpc.init()
    eccmap.init()
    my_keys.init()
    # if sys.argv.count('--twisted'):
    #     from twisted.python import log as twisted_log
    #     twisted_log.startLogging(MyTwistedOutputLog(), setStdout=0)
    # import twisted.python.failure as twisted_failure
    # twisted_failure.startDebugMode()
    # twisted_log.defaultObserver.stop()
    # if settings.getDebugLevel() > 10:
    #     defer.setDebugging(True)
    if settings.enableMemoryProfile():
        try:
            from guppy import hpy  # @UnresolvedImport
            hp = hpy()
            hp.setrelheap()
            if _Debug:
                lg.out(_DebugLevel, 'hp.heap():\n' + str(hp.heap()))
                lg.out(_DebugLevel, 'hp.heap().byrcs:\n' + str(hp.heap().byrcs))
                lg.out(_DebugLevel, 'hp.heap().byvia:\n' + str(hp.heap().byvia))
        except:
            if _Debug:
                lg.dbg(_DebugLevel, 'guppy package is not installed')
    if _Debug:
        lg.dbg(_DebugLevel, 'all local modules are initialized, ready to start the engine')


def init_interfaces():
    from bitdust.main import settings
    # if settings.enableFTPServer():
    #     try:
    #         from bitdust.interface import ftp_server
    #         ftp_server.init()
    #     except:
    #         lg.exc()
    if settings.enableAPIAuthSecret():
        current_secret = local_fs.ReadTextFile(settings.APISecretFile())
        if not current_secret:
            new_secret = cipher.generate_secret_text(10)
            local_fs.WriteTextFile(settings.APISecretFile(), new_secret)
            lg.info('generated new API auth secret text and stored in %r' % settings.APISecretFile())
    if settings.enableRESTHTTPServer():
        try:
            from bitdust.interface import api_rest_http_server
            api_rest_http_server.init(port=settings.getRESTHTTPServerPort())
        except:
            lg.exc()
    if settings.enableWebSocketServer():
        try:
            from bitdust.interface import api_web_socket
            api_web_socket.init(port=settings.getWebSocketServerPort())
        except:
            lg.exc()


def init_services():
    driver.init()
    return driver.start()


#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _Initializer
    if _Initializer is None:
        _Initializer = Initializer(
            name='initializer',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        if kwargs.get('use_reactor', True):
            _Initializer.automat(event, *args, **kwargs)
        else:
            _Initializer.event(event, *args, **kwargs)
    return _Initializer


def Destroy():
    """
    Destroy initializer() automat and remove its instance from memory.
    """
    global _Initializer
    if _Initializer is None:
        return
    _Initializer.destroy()
    del _Initializer
    _Initializer = None


class Initializer(automat.Automat):
    """
    A class to execute start up operations to launch BitDust software.
    """

    fast = True

    def init(self):
        self.flagCmdLine = False
        self.flagGUI = False
        self.is_installed = None

    def A(self, event, *args, **kwargs):
        from bitdust.main import installer
        from bitdust.main import shutdowner
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'run':
                self.state = 'LOCAL'
                shutdowner.A('init')
                self.doInitLocal(*args, **kwargs)
                self.flagCmdLine = False
            elif event == 'run-cmd-line-register':
                self.state = 'INSTALL'
                shutdowner.A('init')
                self.flagCmdLine = True
                installer.A('register-cmd-line', *args, **kwargs)
                shutdowner.A('ready')
            elif event == 'run-cmd-line-recover':
                self.state = 'INSTALL'
                shutdowner.A('init')
                self.flagCmdLine = True
                installer.A('recover-cmd-line', *args, **kwargs)
                shutdowner.A('ready')
        #---LOCAL---
        elif self.state == 'LOCAL':
            if event == 'init-local-done' and not self.isInstalled(*args, **kwargs) and self.isGUIPossible(*args, **kwargs):
                self.state = 'INSTALL'
                installer.A('init')
                shutdowner.A('ready')
                self.doInitInterfaces(*args, **kwargs)
                self.doShowGUI(*args, **kwargs)
                self.doUpdate(*args, **kwargs)
            elif (event == 'shutdowner.state' and args[0] == 'FINISHED'):
                self.state = 'STOPPING'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'init-local-done' and ((not self.isInstalled(*args, **kwargs) and not self.isGUIPossible(*args, **kwargs)) or self.isInstalled(*args, **kwargs)):
                self.state = 'INTERFACES'
                shutdowner.A('ready')
                self.doInitInterfaces(*args, **kwargs)
        #---MODULES---
        elif self.state == 'MODULES':
            if event == 'init-modules-done':
                self.state = 'READY'
                self.doUpdate(*args, **kwargs)
                self.doShowGUI(*args, **kwargs)
            elif (event == 'shutdowner.state' and args[0] == 'FINISHED'):
                self.state = 'EXIT'
                self.doDestroyMe(*args, **kwargs)
        #---INSTALL---
        elif self.state == 'INSTALL':
            if not self.flagCmdLine and (event == 'installer.state' and args[0] == 'DONE'):
                self.state = 'STOPPING'
                shutdowner.A('stop', 'restartnshow')
            elif self.flagCmdLine and (event == 'installer.state' and args[0] == 'DONE'):
                self.state = 'STOPPING'
                shutdowner.A('stop', 'exit')
            elif (event == 'shutdowner.state' and args[0] == 'FINISHED'):
                self.state = 'EXIT'
                self.doDestroyMe(*args, **kwargs)
        #---READY---
        elif self.state == 'READY':
            if (event == 'shutdowner.state' and args[0] == 'FINISHED'):
                self.state = 'EXIT'
                self.doDestroyMe(*args, **kwargs)
        #---STOPPING---
        elif self.state == 'STOPPING':
            if (event == 'shutdowner.state' and args[0] == 'FINISHED'):
                self.state = 'EXIT'
                self.doUpdate(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---EXIT---
        elif self.state == 'EXIT':
            pass
        #---SERVICES---
        elif self.state == 'SERVICES':
            if event == 'init-services-done':
                self.state = 'MODULES'
                self.doInitModules(*args, **kwargs)
                shutdowner.A('ready')
            elif (event == 'shutdowner.state' and args[0] == 'FINISHED'):
                self.state = 'EXIT'
                self.doDestroyMe(*args, **kwargs)
        #---INTERFACES---
        elif self.state == 'INTERFACES':
            if event == 'init-interfaces-done':
                self.state = 'SERVICES'
                self.doInitServices(*args, **kwargs)
            elif (event == 'shutdowner.state' and args[0] == 'FINISHED'):
                self.state = 'EXIT'
                self.doDestroyMe(*args, **kwargs)
        return None

    def isInstalled(self, *args, **kwargs):
        if self.is_installed is None:
            self.is_installed = self._check_install()
        return self.is_installed

    def isGUIPossible(self, *args, **kwargs):
        return bpio.isGUIpossible()

    def doUpdate(self, *args, **kwargs):
        pass

    def doInitLocal(self, *args, **kwargs):
        self.flagGUI = args[0].strip() == 'show'
        if _Debug:
            lg.out(_DebugLevel, 'initializer.doInitLocal flagGUI=%s' % self.flagGUI)
        init_local()
        if bpio.Android():
            self.automat('init-local-done')
        else:
            reactor.callWhenRunning(self.automat, 'init-local-done')  # @UndefinedVariable

    def doInitServices(self, *args, **kwargs):
        """
        Action method.
        """
        if bpio.Android():
            lg.close_intercepted_log_file()
            from android.storage import app_storage_path  # @UnresolvedImport
            lg.open_intercepted_log_file(os.path.join(app_storage_path(), '.bitdust', 'logs', 'android.log'), mode='a')
            if _Debug:
                lg.dbg(_DebugLevel, 'log file "android.log" re-opened')
        if _Debug:
            lg.out(_DebugLevel, 'initializer.doInitServices')
        d = init_services()
        d.addBoth(lambda x: self.automat('init-services-done'))

    def doInitInterfaces(self, *args, **kwargs):
        if _Debug:
            lg.out(_DebugLevel, 'initializer.doInitInterfaces')
        init_interfaces()
        reactor.callLater(0, self.automat, 'init-interfaces-done')  # @UndefinedVariable

    def doInitModules(self, *args, **kwargs):
        if _Debug:
            lg.out(_DebugLevel, 'initializer.doInitModules')
        self._init_auto_update()
        reactor.callLater(0, self.automat, 'init-modules-done')  # @UndefinedVariable

    def doShowGUI(self, *args, **kwargs):
        if _Debug:
            lg.out(_DebugLevel, 'initializer.doShowGUI')
        try:
            from bitdust.system.tray_icon import USE_TRAY_ICON
        except:
            USE_TRAY_ICON = False
            lg.exc()
        if USE_TRAY_ICON:
            from bitdust.system import tray_icon
            tray_icon.SetControlFunc(self._on_tray_icon_command)
        # TODO: raise up electron window ?

    def doDestroyMe(self, *args, **kwargs):
        global _Initializer
        try:
            from bitdust.system.tray_icon import USE_TRAY_ICON
        except:
            USE_TRAY_ICON = False
        if USE_TRAY_ICON:
            from bitdust.system import tray_icon
            tray_icon.SetControlFunc(None)
        del _Initializer
        _Initializer = None
        self.destroy()

    #------------------------------------------------------------------------------

    def _check_install(self):
        """
        Return True if Private Key and local identity files exists and both is
        valid.
        """
        if _Debug:
            lg.out(_DebugLevel, 'initializer._check_install')
        from bitdust.main import settings
        from bitdust.userid import identity
        from bitdust.crypt import key
        keyfilename = settings.KeyFileName()
        keyfilenamelocation = settings.KeyFileNameLocation()
        if os.path.isfile(keyfilenamelocation):
            keyfilename_custom = bpio.ReadTextFile(keyfilenamelocation).strip()
            if os.path.isfile(keyfilename_custom):
                keyfilename = keyfilename_custom
        idfilename = settings.LocalIdentityFilename()
        if not os.path.exists(keyfilename) or not os.path.exists(idfilename):
            if _Debug:
                lg.out(_DebugLevel, 'initializer._check_install local key or local id not exists')
            return False
        current_key = bpio.ReadTextFile(keyfilename)
        current_id = bpio.ReadTextFile(idfilename)
        if not current_id:
            if _Debug:
                lg.out(_DebugLevel, 'initializer._check_install local identity is empty ')
            return False
        if not current_key:
            if _Debug:
                lg.out(_DebugLevel, 'initializer._check_install private key is empty ')
            return False
        try:
            key.InitMyKey()
        except:
            if _Debug:
                lg.out(_DebugLevel, 'initializer._check_install fail loading private key ')
            return False
        try:
            ident = identity.identity(xmlsrc=current_id)
        except:
            if _Debug:
                lg.out(_DebugLevel, 'initializer._check_install fail init local identity ')
            return False
        if not ident.isCorrect():
            lg.err('local identity is not correct !!!')
            return False
        try:
            res = ident.Valid()
        except:
            if _Debug:
                lg.out(_DebugLevel, 'failed to validate local identity')
            return False
        if not res:
            lg.err('local identity is not valid !!!')
            return False
        if _Debug:
            lg.out(_DebugLevel, 'initializer._check_install SUCCESS!!!')
        return True

    def _on_software_code_updated(self, evt):
        if _Debug:
            lg.out(_DebugLevel, 'initializer._on_software_code_updated will RESTART BitDust now! "source-code-fetched" event received')
        if False:
            # TODO: add checks to prevent restart if any important jobs running at the moment
            return
        if False:
            # TODO: add an option to the settings
            return
        from bitdust.main import shutdowner
        shutdowner.A('stop', 'restart')

    def _init_auto_update(self):
        if _Debug:
            lg.out(_DebugLevel, 'initializer._init_modules')
        from bitdust.updates import git_proc
        from bitdust.main import events
        git_proc.init()
        events.add_subscriber(self._on_software_code_updated, 'source-code-fetched')

    def _on_tray_icon_command(self, cmd):
        if _Debug:
            lg.out(_DebugLevel, 'initializer._on_tray_icon_command : [%s]' % cmd)
        try:
            from bitdust.main import shutdowner
            if cmd == 'exit':
                shutdowner.A('stop', 'exit')

            elif cmd == 'restart':
                shutdowner.A('stop', 'restart')

            elif cmd == 'reconnect':
                from bitdust.p2p import network_connector
                if driver.is_on('service_network'):
                    network_connector.A('reconnect')

            elif cmd == 'show':
                # TODO: raise up electron window ?
                pass

            elif cmd == 'sync':
                try:
                    from bitdust.updates import git_proc
                    from bitdust.system import tray_icon

                    def _sync_callback(result):
                        if result == 'error':
                            tray_icon.draw_icon('error')
                            reactor.callLater(5, tray_icon.restore_icon)  # @UndefinedVariable
                            return
                        elif result == 'source-code-fetched':
                            tray_icon.draw_icon('updated')
                            reactor.callLater(5, tray_icon.restore_icon)  # @UndefinedVariable
                            return
                        tray_icon.restore_icon()

                    tray_icon.draw_icon('sync')
                    git_proc.sync(_sync_callback, update_method='reset')
                except:
                    lg.exc()

            elif cmd == 'hide':
                pass

            elif cmd == 'toolbar':
                pass

            else:
                lg.warn('wrong command: ' + str(cmd))
        except:
            lg.exc()


#------------------------------------------------------------------------------


class MyTwistedOutputLog:
    softspace = 0

    def read(self):
        pass

    def write(self, s):
        if _Debug:
            lg.out(_DebugLevel, 'TWISTED: ' + s.strip())

    def flush(self):
        pass

    def close(self):
        pass



#!/usr/bin/python
# child_process.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (child_process.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: child_process.


"""

from __future__ import absolute_import
import os
import sys
import subprocess

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in child_process.py')

from twisted.internet import protocol

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import nonblocking

#------------------------------------------------------------------------------


class ChildProcessProtocol(protocol.ProcessProtocol):
    def __init__(self, name):
        self.name = name

    def errReceived(self, inp):
        for line in inp.splitlines():
            lg.out(2, '[%s]: %s' % (self.name, line))

    def processEnded(self, reason):
        lg.out(2, 'child process [%s] FINISHED' % self.name)


#------------------------------------------------------------------------------


def run(child_name, params=[], base_dir='.', process_protocol=None):
    """
    This is another portable solution to execute a process.
    """
    if bpio.isFrozen() and bpio.Windows():
        progpath = os.path.abspath(os.path.join(base_dir, child_name + '.exe'))
        executable = progpath
        cmdargs = [progpath]
        cmdargs.extend(params)
    else:
        progpath = os.path.abspath(os.path.join(base_dir, child_name + '.py'))
        executable = sys.executable
        cmdargs = [executable, progpath]
        cmdargs.extend(params)
    if not os.path.isfile(executable):
        lg.out(1, 'child_process.run ERROR %s not found' % executable)
        return None
    if not os.path.isfile(progpath):
        lg.out(1, 'child_process.run ERROR %s not found' % progpath)
        return None
    lg.out(6, 'child_process.run: "%s"' % (' '.join(cmdargs)))

    if bpio.Windows():
        from twisted.internet import _dumbwin32proc
        real_CreateProcess = _dumbwin32proc.win32process.CreateProcess  # @UndefinedVariable

        def fake_createprocess(_appName, _commandLine, _processAttributes, _threadAttributes, _bInheritHandles, creationFlags, _newEnvironment, _currentDirectory, startupinfo):
            import win32con  # @UnresolvedImport
            flags = win32con.CREATE_NO_WINDOW
            return real_CreateProcess(_appName, _commandLine, _processAttributes, _threadAttributes, _bInheritHandles, flags, _newEnvironment, _currentDirectory, startupinfo)

        setattr(_dumbwin32proc.win32process, 'CreateProcess', fake_createprocess)

    if process_protocol is None:
        process_protocol = ChildProcessProtocol(child_name)
    try:
        Process = reactor.spawnProcess(process_protocol, executable, cmdargs, path=base_dir)  # @UndefinedVariable
    except:
        lg.out(1, 'child_process.run ERROR executing: %s' % str(cmdargs))
        lg.exc()
        return None

    if bpio.Windows():
        setattr(_dumbwin32proc.win32process, 'CreateProcess', real_CreateProcess)

    lg.out(6, 'child_process.run [%s] pid=%d' % (child_name, Process.pid))
    return Process


def kill_process(process):
    """
    Send signal "KILL" to the given ``process``.
    """
    try:
        process.signalProcess('KILL')
        lg.out(6, 'child_process.kill_process sent signal "KILL" to the process %d' % process.pid)
    except:
        return False
    return True


def kill_child(child_name):
    """
    Search (by "pid") for BitDust child process with name ``child_name`` and
    tries to kill it.
    """
    killed = False
    for pid in bpio.find_process([child_name + '.']):
        bpio.kill_process(pid)
        lg.out(6, 'child_process.kill_child pid %d' % pid)
        killed = True
    return killed


#------------------------------------------------------------------------------


def pipe(cmdargs):
    """
    Execute a process in different way, create a Pipe to do read/write
    operations with child process.

    See ``lib.nonblocking`` module.
    """
    lg.out(6, 'child_process.pipe %s' % str(cmdargs))
    try:
        if bpio.Windows():
            import win32process  # @UnresolvedImport
            p = nonblocking.Popen(
                cmdargs,
                shell=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=False,
                creationflags=win32process.CREATE_NO_WINDOW,
            )
        else:
            p = nonblocking.Popen(
                cmdargs,
                shell=False,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=False,
            )
    except:
        lg.out(1, 'child_process.pipe ERROR executing: %s' + str(cmdargs))
        lg.exc()
        return None
    return p


def detach(cmdargs):
    lg.out(2, 'child_process.detach %s' % str(cmdargs))
    try:
        if bpio.Windows():
            import win32process  # @UnresolvedImport
            p = nonblocking.Popen(
                cmdargs,
                shell=False,
                # stdin=subprocess.PIPE,
                # stdout=subprocess.PIPE,
                # stderr=subprocess.PIPE,
                universal_newlines=False,
                creationflags=win32process.CREATE_NO_WINDOW | win32process.DETACHED_PROCESS,
                close_fds=True,
            )
        else:
            p = nonblocking.Popen(
                cmdargs,
                shell=False,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=False,
                close_fds=True,
            )
    except:
        lg.out(1, 'child_process.detach ERROR executing: %s' + str(cmdargs))
        lg.exc()
        return None
    return p



#!/usr/bin/python
# events.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (events.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: events.

A very simple "event" system, just to show and remember what is goin on.
Also you can subscribe to given event and receive notifications.

TODO: need to store events on the local HDD
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

_EventLogFileEnabled = True
_EventLogUseColors = None

#------------------------------------------------------------------------------

import os
import sys

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in events.py')

from twisted.internet.defer import Deferred  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import utime

#------------------------------------------------------------------------------

MAX_PENDING_EVENTS_PER_CONSUMER = 100

#------------------------------------------------------------------------------

_Subscribers = {}
_ConsumersCallbacks = {}
_EventQueuePerConsumer = {}
_EventsCount = {}

#------------------------------------------------------------------------------


def subscribers():
    global _Subscribers
    return _Subscribers


#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'events.init')
    add_subscriber(push_event)


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'events.shutdown')
    clear_subscribers()
    subscribers().clear()
    event_queue().clear()
    count().clear()


#------------------------------------------------------------------------------


class Event(object):
    def __init__(self, event_id, data=None, created=None):
        self.event_id = event_id
        self.data = data
        self.created = created or utime.utcnow_to_sec1970()

    def __repr__(self):
        return '<Event({})>'.format(self.event_id)


#------------------------------------------------------------------------------


def add_subscriber(subscriber_callback, event_id='*'):
    """
    subscriber_callback(evt)
    """
    if event_id not in subscribers():
        subscribers()[event_id] = []
    subscribers()[event_id].append(subscriber_callback)
    return True


def remove_subscriber(subscriber_callback, event_id='*'):
    removed = False
    if event_id == '*':
        for event_id, subscriber_callbacks in subscribers().items():
            if subscriber_callback in subscriber_callbacks:
                subscribers()[event_id].remove(subscriber_callback)
                removed = True
    else:
        if event_id in subscribers():
            if subscriber_callback in subscribers()[event_id]:
                subscribers()[event_id].remove(subscriber_callback)
                removed = True
    return removed


def clear_subscribers(event_id='*'):
    removed = False
    for _event_id, subscriber_callbacks in subscribers().items():
        if _event_id == event_id or event_id == '*':
            for _cb in list(subscriber_callbacks):
                subscribers()[_event_id].remove(_cb)
                removed = True
    return removed


#------------------------------------------------------------------------------


def dispatch(evt):
    global _EventLogFileEnabled
    global _EventLogUseColors
    global _EventsCount
    handled = 0
    if evt.event_id not in _EventsCount:
        _EventsCount[evt.event_id] = 0
    _EventsCount[evt.event_id] += 1
    if evt.event_id in subscribers():
        for subscriber_callback in subscribers()[evt.event_id]:
            try:
                subscriber_callback(evt)
            except:
                lg.exc()
                continue
            handled += 1
    if '*' in subscribers():
        for subscriber_callback in subscribers()['*']:
            try:
                subscriber_callback(evt)
            except:
                lg.exc()
                continue
            handled += 1
    if _EventLogFileEnabled:
        if _EventLogUseColors is None:
            _EventLogUseColors = os.environ.get('BITDUST_LOG_USE_COLORS', '1') != '0'
        if _EventLogUseColors:
            lg.out(2, '\033[0;49;91m%s\033[0m  %r' % (evt.event_id, str(evt.data)), log_name='event', showtime=True)
        else:
            lg.out(2, '%s  %r' % (evt.event_id, str(evt.data)), log_name='event', showtime=True)
    if _Debug:
        if not handled:
            lg.warn('event "{}" was not handled'.format(evt.event_id))
        else:
            if _EventLogUseColors:
                lg.out(_DebugLevel, '\033[0;49;36mevents.dispatch "{}" was handled by {} subscribers\033[0m'.format(evt.event_id, handled))
            else:
                lg.out(_DebugLevel, 'events.dispatch "{}" was handled by {} subscribers'.format(evt.event_id, handled))
    return handled


def send(event_id, data=None, created=None, fast=False):
    evt = Event(event_id, data=data, created=created)
    if fast:
        dispatch(evt)
    else:
        reactor.callLater(0, dispatch, evt)  # @UndefinedVariable
    return evt


#------------------------------------------------------------------------------


def event_queue():
    global _EventQueuePerConsumer
    return _EventQueuePerConsumer


def consumers_callbacks():
    global _ConsumersCallbacks
    return _ConsumersCallbacks


def consume_events(consumer_id):
    if consumer_id not in consumers_callbacks():
        consumers_callbacks()[consumer_id] = []
    d = Deferred()
    consumers_callbacks()[consumer_id].append(d)
    if _Debug:
        lg.out(_DebugLevel, 'events.consume_events added callback for consumer "%s", %d total callbacks' % (consumer_id, len(consumers_callbacks()[consumer_id])))
    reactor.callLater(0, pop_event)  # @UndefinedVariable
    return d


def push_event(event_object):
    for consumer_id in consumers_callbacks().keys():
        if consumer_id not in event_queue():
            event_queue()[consumer_id] = []
        event_queue()[consumer_id].append({
            'type': 'event',
            'id': event_object.event_id,
            'data': event_object.data,
            'time': event_object.created,
        })
        if _Debug:
            lg.out(_DebugLevel, 'events.push_event "%s" for consumer "%s", %d pending events' % (event_object.event_id, consumer_id, len(event_queue()[consumer_id])))
    reactor.callLater(0, pop_event)  # @UndefinedVariable


def pop_event():
    for consumer_id in list(consumers_callbacks().keys()):
        if consumer_id not in event_queue() or len(event_queue()[consumer_id]) == 0:
            continue
        registered_callbacks = consumers_callbacks()[consumer_id]
        pending_messages = event_queue()[consumer_id]
        if len(registered_callbacks) == 0 and len(pending_messages) > MAX_PENDING_EVENTS_PER_CONSUMER:
            consumers_callbacks().pop(consumer_id)
            event_queue().pop(consumer_id)
            if _Debug:
                lg.out(_DebugLevel, 'events.pop_event STOPPED consumer "%s", too many pending messages but no callbacks' % consumer_id)
            continue
        for consumer_callback in registered_callbacks:
            if not consumer_callback:
                if _Debug:
                    lg.out(_DebugLevel, 'events.pop_event %d events waiting consuming by "%s", no callback yet' % (len(event_queue()[consumer_id]), consumer_id))
                continue
            if consumer_callback.called:
                if _Debug:
                    lg.out(_DebugLevel, 'events.pop_event %d events waiting consuming by "%s", callback state is "called"' % (len(event_queue()[consumer_id]), consumer_id))
                continue
            consumer_callback.callback(pending_messages)
            event_queue()[consumer_id] = []
            if _Debug:
                lg.out(_DebugLevel, 'events.pop_event %d events consumed by "%s"' % (len(pending_messages), consumer_id))
        consumers_callbacks()[consumer_id] = []


#------------------------------------------------------------------------------


def count(event_id=None):
    global _EventsCount
    if 'my-identity-rotated' not in _EventsCount:
        _EventsCount['my-identity-rotated'] = 0
    if 'identity-rotated' not in _EventsCount:
        _EventsCount['identity-rotated'] = 0
    if event_id is None:
        return _EventsCount
    return _EventsCount.get(event_id, 0)



#!/usr/bin/python
# service_keys_registry.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_keys_registry.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_keys_registry
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return KeysRegistryService()


class KeysRegistryService(LocalService):

    service_name = 'service_keys_registry'
    config_path = 'services/keys-registry/enabled'

    def dependent_on(self):
        return [
            'service_p2p_notifications',
        ]

    def start(self):
        from bitdust.transport import callback
        from bitdust.main import listeners
        from bitdust.crypt import my_keys
        from bitdust.access import key_ring
        key_ring.init()
        callback.add_outbox_callback(self._on_outbox_packet_sent)
        callback.append_inbox_callback(self._on_inbox_packet_received)
        if listeners.is_populate_required('key'):
            # listeners.populate_later().remove('key')
            my_keys.populate_keys()
        return True

    def stop(self):
        from bitdust.transport import callback
        from bitdust.access import key_ring
        callback.remove_inbox_callback(self._on_inbox_packet_received)
        callback.remove_outbox_callback(self._on_outbox_packet_sent)
        key_ring.shutdown()
        return True

    def request(self, json_payload, newpacket, info):
        # TODO: work in progress
        # from bitdust.main import events
        from bitdust.p2p import p2p_service
        # events.send('key-registry-request', data=dict(idurl=newpacket.OwnerID))
        return p2p_service.SendAck(newpacket, 'accepted')

    def _on_outbox_packet_sent(self, pkt_out):
        from bitdust.p2p import commands
        if pkt_out.outpacket.Command == commands.Key():
            # TODO: work in progress : need to store history of all keys transfers
            return True
        return False

    def _on_inbox_packet_received(self, newpacket, info, status, error_message):
        from bitdust.p2p import commands
        from bitdust.access import key_ring
        if newpacket.Command == commands.Key():
            # TODO: work in progress : need to store history of all keys transfers
            return key_ring.on_key_received(newpacket, info, status, error_message)
        elif newpacket.Command == commands.AuditKey():
            return key_ring.on_audit_key_received(newpacket, info, status, error_message)
        return False



#!/usr/bin/env python
# customer_contract_executor.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (customer_contract_executor.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: customer_contract_executor
.. role:: red

EVENTS:
    * :red:`coin-failed`
    * :red:`coin-published`
    * :red:`coins-received`
    * :red:`contract-extended`
    * :red:`contract-finished`
    * :red:`init`
    * :red:`recheck`
    * :red:`request-failed`
    * :red:`shutdown`
    * :red:`supplier-coin-mined`
    * :red:`supplier-failed`
    * :red:`time-to-pay`
    * :red:`timer-1min`
    * :red:`timer-30min`
    * :red:`timer-30sec`
    * :red:`timer-5min`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import nameurl

from bitdust.userid import my_id

#------------------------------------------------------------------------------

_ActiveCustomerContracts = dict()  # provides CustomerContractExecutor object by supplier idurl

#------------------------------------------------------------------------------


def all_contracts():
    global _ActiveCustomerContracts
    return _ActiveCustomerContracts


def init_contract(supplier_idurl):
    global _ActiveCustomerContracts
    if supplier_idurl in _ActiveCustomerContracts:
        lg.warn('contract with supplier %s already started' % supplier_idurl)
        return _ActiveCustomerContracts[supplier_idurl]
    cce = CustomerContractExecutor(supplier_idurl)
    cce.automat('init')
    _ActiveCustomerContracts[supplier_idurl] = cce
    return _ActiveCustomerContracts[supplier_idurl]


def shutdown_contract(supplier_idurl):
    global _ActiveCustomerContracts
    if supplier_idurl not in _ActiveCustomerContracts:
        lg.warn('no contract started for given supplier')
        return False
    _ActiveCustomerContracts[supplier_idurl].automat('shutdown')
    _ActiveCustomerContracts.pop(supplier_idurl)
    return True


def get_contract(supplier_idurl):
    global _ActiveCustomerContracts
    if supplier_idurl not in _ActiveCustomerContracts:
        return None
    return _ActiveCustomerContracts[supplier_idurl]


def recheck_contract(supplier_idurl):
    contract_executor = get_contract(supplier_idurl)
    if contract_executor is None:
        contract_executor = init_contract(supplier_idurl)
    contract_executor.automat('recheck')


#------------------------------------------------------------------------------


class CustomerContractExecutor(automat.Automat):
    """
    This class implements all the functionality of the ``customer_contract_executor()`` state machine.
    """

    timers = {
        'timer-30min': (1800, ['SUPPLIER_COIN?']),
        'timer-1min': (60, ['READ_COINS?']),
        'timer-30sec': (30.0, ['EXTENSION?']),
        'timer-5min': (300, ['SUPERVISE']),
    }

    def __init__(self, supplier_idurl):
        self.supplier_idurl = supplier_idurl
        name = 'customer_executor_%s' % nameurl.GetName(self.supplier_idurl)
        automat.Automat.__init__(self, name, 'AT_STARTUP', _DebugLevel, _Debug)

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <https://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'READ_COINS?'
                self.doRequestCoins(*args, **kwargs)
        #---READ_COINS?---
        elif self.state == 'READ_COINS?':
            if event == 'coins-received':
                self.state = 'SUPERVISE'
                self.doSchedulePayment(*args, **kwargs)
                self.doSuperviseSupplier(*args, **kwargs)
            elif event == 'timer-1min' or event == 'request-failed':
                self.state = 'UNCLEAR'
        #---SUPERVISE---
        elif self.state == 'SUPERVISE':
            if event == 'timer-5min':
                self.doSuperviseSupplier(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doCancelPayment(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'time-to-pay' and self.isOkToContinue(*args, **kwargs):
                self.state = 'MY_COIN!'
                self.doSendNextCoin(*args, **kwargs)
            elif event == 'time-to-pay' and not self.isOkToContinue(*args, **kwargs):
                self.state = 'MY_COIN!'
                self.doSendLastCoin(*args, **kwargs)
            elif event == 'supplier-failed':
                self.state = 'MY_COIN!'
                self.doCancelPayment(*args, **kwargs)
                self.doSendLastCoin(*args, **kwargs)
                self.doReplaceSupplier(*args, **kwargs)
        #---SUPPLIER_COIN?---
        elif self.state == 'SUPPLIER_COIN?':
            if event == 'supplier-coin-mined':
                self.state = 'EXTENSION?'
                self.doRequestService(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'timer-30min':
                self.state = 'UNCLEAR'
        #---EXTENSION?---
        elif self.state == 'EXTENSION?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'contract-extended':
                self.state = 'SUPERVISE'
                self.doSchedulePayment(*args, **kwargs)
                self.doSuperviseSupplier(*args, **kwargs)
            elif event == 'contract-finished' or event == 'timer-30sec':
                self.state = 'FINISHED'
                self.doSendFinishCoin(*args, **kwargs)
                self.doReplaceSupplier(*args, **kwargs)
        #---FINISHED---
        elif self.state == 'FINISHED':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        #---MY_COIN!---
        elif self.state == 'MY_COIN!':
            if event == 'coin-published' and not self.isChainClosed(*args, **kwargs):
                self.state = 'SUPPLIER_COIN?'
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'coin-failed':
                self.state = 'UNCLEAR'
            elif event == 'coin-published' and self.isChainClosed(*args, **kwargs):
                self.state = 'FINISHED'
                self.doSendFinishCoin(*args, **kwargs)
                self.doReplaceSupplier(*args, **kwargs)
        #---UNCLEAR---
        elif self.state == 'UNCLEAR':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'recheck':
                self.state = 'READ_COINS?'
                self.doRequestCoins(*args, **kwargs)
        return None

    def isOkToContinue(self, *args, **kwargs):
        """
        Condition method.
        """

    def isChainClosed(self, *args, **kwargs):
        """
        Condition method.
        """

    def doRequestCoins(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.coins import contract_chain_node
        contract_chain_node.get_coins_by_chain(
            chain='supplier_customer',
            provider_idurl=self.supplier_idurl,
            consumer_idurl=my_id.getIDURL(),
        ).addCallbacks(
            self._on_coins_received,
            self._on_coins_failed,
        )

    def doSuperviseSupplier(self, *args, **kwargs):
        """
        Action method.
        """
        #TODO: create supplier_supervisor() machine

    def doSchedulePayment(self, *args, **kwargs):
        """
        Action method.
        """

    def doCancelPayment(self, *args, **kwargs):
        """
        Action method.
        """

    def doRequestService(self, *args, **kwargs):
        """
        Action method.
        """

    def doReplaceSupplier(self, *args, **kwargs):
        """
        Action method.
        """

    def doSendNextCoin(self, *args, **kwargs):
        """
        Action method.
        """

    def doSendLastCoin(self, *args, **kwargs):
        """
        Action method.
        """

    def doSendSuccessCoin(self, *args, **kwargs):
        """
        Action method.
        """

    def doSendFailedCoin(self, *args, **kwargs):
        """
        Action method.
        """

    def doSendFinishCoin(self, *args, **kwargs):
        """
        Action method.
        """

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.destroy()

    def _on_coins_received(self, coins):
        if _Debug:
            lg.out(_DebugLevel, 'customer_contract_executor._on_coins_received %s' % coins)
        if not coins:
            # self.automat('')
            return

    def _on_coins_failed(self, fails):
        if _Debug:
            lg.warn(str(fails))
